unit AsBarCode_interface;

{<
  BarCode_interface contain interface function for generating of Barcodes,
  used by RTFReport and XMLReport but can be used independently by user too.

  Full functional demo.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2006)
  @lastmod(Apr 2014)
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics;

type
  { list of supported Barcode types }
  TBarcodeType = (bcCode_2_5_interleaved, bcCode_2_5_industrial, bcCode_2_5_matrix, bcCode39, bcCode39Extended, bcCode128A, bcCode128B, bcCode128C, bcCode93,
    bcCode93Extended, bcCodeMSI, bcCodePostNet, bcCodeCodabar, bcCodeEAN8, bcCodeEAN13, bcCodeUPC_A, bcCodeUPC_E0, bcCodeUPC_E1, bcCodeUPC_Supp2,
    bcCodeUPC_Supp5, bcCodeEAN128A, bcCodeEAN128B, bcCodeEAN128C, bcTypeUndefined);

  { list of possible BarCode text positions }
  TShowTextPosition = (stpTopLeft, stpTopRight, stpTopCenter, stpBottomLeft, stpBottomRight, stpBottomCenter);

  { record for definition of one BarCode type }
  TBCdata = record
    Name: string;
    num: boolean;
  end;

const
  { array with definitions of all barcodes }
  BCdata: array [bcCode_2_5_interleaved .. bcTypeUndefined] of TBCdata = ((Name: 'Code_2_5_interleaved'; num: true), (Name: 'Code_2_5_industrial'; num: true),
    (Name: 'Code_2_5_matrix'; num: true), (Name: 'Code39'; num: false), (Name: 'Code39Extended'; num: false), (Name: 'Code128A'; num: false),
    (Name: 'Code128B'; num: false), (Name: 'Code128C'; num: true), (Name: 'Code93'; num: false), (Name: 'Code93Extended'; num: false),
    (Name: 'CodeMSI'; num: true), (Name: 'CodePostNet'; num: true), (Name: 'CodeCodabar'; num: false), (Name: 'CodeEAN8'; num: true),
    (Name: 'CodeEAN13'; num: true), (Name: 'CodeUPC_A'; num: true), (Name: 'CodeUPC_E0'; num: true), (Name: 'CodeUPC_E1'; num: true),
    (Name: 'CodeUPC_Supp2'; num: true), (Name: 'CodeUPC_Supp5'; num: true), (Name: 'CodeEAN128A'; num: false), (Name: 'CodeEAN128B'; num: false),
    (Name: 'CodeEAN128C'; num: true), (Name: 'Undefined'; num: true));

  BCCheckSumType = [bcCodeEAN8, bcCodeEAN13, bcCodeUPC_A];

  { load list of BarCode names }
function GetBarCodeNames(L: TStrings; IncludeUndefined: boolean = false): integer;
{ return BarCode type according BarCode name }
function GetBarCodeType(Name: string): TBarcodeType;

{ draw selected BarCode to canvas, Rect return real size of generated barcode,
  if LineWidth=0 then it is calculated from Rect and returned }
procedure DrawBarCode(BarCode: string; BarCodeType: TBarcodeType; Canvas: Pointer; var Rect: TRect; var LineWidth: integer; Font: Pointer;
  AutoFontSize: boolean; ShowText: boolean; ShowTextPosition: TShowTextPosition; UseCheckSum: boolean = true); StdCall;

implementation

uses AsBarcode;

function GetBarCodeNames(L: TStrings; IncludeUndefined: boolean = false): integer;
var
  bctype: TBarcodeType;
begin
  L.Clear;
  for bctype:= Low(TBarcodeType) to High(TBarcodeType) do
    if (bctype <> bcTypeUndefined) or IncludeUndefined then
      L.Add(BCdata[bctype].Name);
  Result:=L.Count;
end;

function GetBarCodeType(Name: string): TBarcodeType;
var
  bctype: TBarcodeType;
begin
  for bctype:= Low(TBarcodeType) to High(TBarcodeType) do
    if SameText(Name, BCdata[bctype].Name) then
    begin
      Result:=bctype;
      Exit;
    end;
  Result:=bcTypeUndefined;
end;

procedure DrawBarCode(BarCode: string; BarCodeType: TBarcodeType; Canvas: Pointer; var Rect: TRect; var LineWidth: integer; Font: Pointer;
  AutoFontSize: boolean; ShowText: boolean; ShowTextPosition: TShowTextPosition; UseCheckSum: boolean = true); StdCall;
var
  bc: TAsBarCode;
begin
  bc:=TAsBarCode.Create(nil);
  try
    bc.BarCode:=BarCode;
    bc.BarCodeType:=BarCodeType;
    bc.UseCheckSum:=UseCheckSum;
    bc.ShowText:=ShowText;
    bc.ShowTextFont.Assign(Font);
    bc.AutoFontSize:=AutoFontSize;
    bc.ShowTextPosition:=ShowTextPosition;

    // do not change order of size settings
    if LineWidth <> 0 then
      bc.Modul:=LineWidth
    else
      LineWidth:=bc.Modul;
    bc.Top:=Rect.Top;
    bc.Left:=Rect.Left;
    bc.Height:=Rect.Bottom - Rect.Top;
    bc.Width:=Rect.Right - Rect.Left;
    Rect.Right:=Rect.Left + bc.Width;

    bc.DrawBarCode(Canvas);
  finally
    bc.Free;
  end;
end;

end.
