unit rdbtool_MDB;

{<
  rdbtool_MDB is set functions for working with MDB database file
  by using of DAO; can be used for updating of database structure
  by application.


  Date: Apr 2008
  Author: Rosi (http://www.rosinsky.cz/delphi)
  Full functional demo
}

interface

uses Windows, Classes, Controls, SysUtils, ActiveX, comobj, dao2000,
  Variants;

type
  TFieldRec = record
    FieldName: string;
    FieldType: integer;
    FieldSize: integer;
    Attributes: integer;
    Required: boolean;
    AllowZeroLength: boolean;
    DefaultValue: OleVariant;
    ForeignName: string;
    ValidationText: string;
    ValidationRule: string;
  end;

  TFieldRecArray = array of TFieldRec;

  TIndexRec = record
    IndexName: string;
    Primary, Unique, Required: boolean;
    FieldName: string;
  end;

  TIndexRecArray = array of TIndexRec;

  TQueryDef = record
    Name: string;
    SQL: string;
  end;

  TQueryDefArray = array of TQueryDef;

  TRelationRec = record
    Name, Table, ForeignTable: string;
    Attributes: integer;
    Fields: TFieldRecArray;
  end;

  TRelationRecArray = array of TRelationRec;

function IsAccess97(DBName: string): boolean;
function OpenMDB(DBName, Pwd: string): Database;
procedure CompactMDBEx(SrcName, DstName, OldPwd, NewPwd: string);
function CompactMDB(DBName, Pwd: string): boolean;
function ReadMDBUsers(DBName: string; List: TStrings): boolean;

procedure GetTableList(DBName, Pwd: string; List: TStrings); overload;
procedure GetTableList(db: Database; List: TStrings); overload;

function GetTableFields(DBName, Pwd, TableName: string; List: TStrings): TFieldRecArray; overload;
function GetTableFields(db: Database; TableName: string; List: TStrings): TFieldRecArray; overload;

function GetTableIndexes(DBName, Pwd, TableName: string; NotRelations: boolean; List: TStrings): TIndexRecArray; overload;
function GetTableIndexes(db: Database; TableName: string; NotRelations: boolean; List: TStrings): TIndexRecArray; overload;

function CreateTable(db: Database; TableName: string; FieldRecArray: TFieldRecArray; IndexRecArray: TIndexRecArray): TableDef;
procedure DropTable(db: Database; TableName: string);
procedure DeleteIndexes(db: Database; TableName: string; LeftPrimaryKey: boolean);
procedure RenameTable(db: Database; SrcTableName, DstTableName: string);

function GetQueryDefs(DBName, Pwd: string; List: TStrings): TQueryDefArray; overload;
function GetQueryDefs(db: Database; List: TStrings): TQueryDefArray; overload;
procedure ClearQueryDefs(db: Database);
function CreateQueryDef(db: Database; QDef: TQueryDef): QueryDef;
procedure CreateQueryDefs(db: Database; QDefArr: TQueryDefArray);

function GetRelations(DBName, Pwd: string; List: TStrings): TRelationRecArray; overload;
function GetRelations(db: Database; List: TStrings): TRelationRecArray; overload;
procedure CreateRelations(db: Database; RelationRecArray: TRelationRecArray);
procedure ClearRelations(db: Database);

procedure RegDAO350;

implementation

uses rtool, rstring, rdlg;

// database
function IsAccess97(DBName: string): boolean;
var
  fi: file of byte;
  i: integer;
  b: byte;
begin
  Result:=false;
  AssignFile(fi, DBName);
  Reset(fi);
  try
    i:=0;
    try
      repeat
        Read(fi, b);
        Inc(i);
        if i = $15 then
        begin
          Result:=b <> 1;
          Break;
        end;
      until Eof(fi);
    except
      raise Exception.Create(sReadDataErr + #13 + Format(sFile, [DBName]));
    end;
  finally
    CloseFile(fi);
  end;
end;

procedure RegDAO350;
var
  s, p, dll, reg: string;
begin
  p:=GetAppPath;
  dll:=p + 'DAO350.dll';
  reg:=p + 'DAO350.reg';
  if not FileExists(dll) then
    raise Exception.Create(Format(sFileNotFound, [dll]));
  if not FileExists(reg) then
    raise Exception.Create(Format(sFileNotFound, [reg]));

  ShowInfo(sInitialize, nil);
  s:=Format('/u /s "%s"', [dll]);
  RunFile('Regsvr32.exe', s, '', SW_SHOW);
  Sleep(1000);
  s:=Format('/s "%s"', [dll]);
  RunFile('Regsvr32.exe', s, '', SW_SHOW);
  Sleep(1000);
  s:=Format('/s "%s"', [reg]);
  RunFile('Regedit.exe', s, '', SW_SHOW);
  Sleep(1000);
  ShowInfo('', nil);
end;

function OpenMDB(DBName, Pwd: string): Database;
var
  db: Database;
  dbEngine: _dbengine;
begin
  if Pwd <> '' then
    Pwd:=';pwd=' + Pwd;

  begin
    dbEngine:=CreateComObject(dao2000.CLASS_DBEngine) as _dbengine;
    db:=dbEngine.OpenDatabase(DBName, dbDriverNoPrompt, false, Pwd)
  end;
  Result:=db;
end;

procedure CompactMDBEx(SrcName, DstName, OldPwd, NewPwd: string);
const
  dbLangGeneral = ';LANGID=0x0409;CP=1252;COUNTRY=0';
var
  dbEngine: _dbengine;
begin
  if OldPwd <> '' then
    OldPwd:=';pwd=' + OldPwd;
  if NewPwd <> '' then
    NewPwd:=';pwd=' + NewPwd;
  begin
    dbEngine:=CreateComObject(dao2000.CLASS_DBEngine) as _dbengine;
    dbEngine.CompactDatabase(SrcName, DstName, NewPwd, dbVersion40, OldPwd);
    // dbEngine.CompactDatabase(SrcName,DstName,null,null,OldPwd);
  end;
end;

function CompactMDB(DBName, Pwd: string): boolean;
var
  Dest, Old: string;
  procedure DelFile(f: string);
  begin
    if FileExists(f) then
      if not DeleteFile(f) then
        raise Exception.Create('Cannot delete file: ' + f);
  end;

begin
  Result:=false;
  try
    Dest:=ExtractFilePath(DBName) + '_compact_.mdb';
    Old:=ExtractFilePath(DBName) + '_old_.mdb';
    DelFile(Dest);
    DelFile(Old);
    CompactMDBEx(DBName, Dest, Pwd, Pwd);
    if not RenameFile(DBName, Old) then
      raise Exception.Create('Cannot rename file: ' + DBName);
    if not RenameFile(Dest, DBName) then
      raise Exception.Create('Cannot rename file: ' + Dest);
    DelFile(Dest);
    DelFile(Old);
    Result:=true;
  except
    on E: Exception do
      ShowErr(sCompressDataErr, E, false);
  end;
  NCursor;
end;

function ReadMDBUsers(DBName: string; List: TStrings): boolean;
const
  DLLFileName = 'MSLDBUSR.dll';
type
  TLDBUser_GetUsers = function(var UserBuffer: PSafeArray; DatabaseName: PChar; Options: integer): integer; stdcall;
  TLDBUser_GetError = function(ErrorNo: integer): PChar; stdcall;
var
  DLLHandle: THandle;
  LDBUser_GetUsers: TLDBUser_GetUsers;
  LDBUser_GetError: TLDBUser_GetError;
  UserCount: integer;
  UserBuffer: Variant;
  Options: integer;
  i: integer;
begin
  Result:=false;
  List.Clear;
  WCursor;
  try
    DLLHandle:=LoadLibrary(DLLFileName);
    if DLLHandle <= 0 then
      List.Add(Format(sDLLNotFound, [DLLFileName]))
    else
      try
        @LDBUser_GetUsers:=GetProcAddress(DLLHandle, 'LDBUser_GetUsers');
        @LDBUser_GetError:=GetProcAddress(DLLHandle, 'LDBUser_GetError');
        if (@LDBUser_GetUsers = nil) or (@LDBUser_GetError = nil) then
          List.Add(Format(sDLLNotFound, [DLLFileName]))
        else
        begin
          UserBuffer:=VarArrayCreate([1, 1], varOleStr);
          // options: 1=all users 2=connected users
          // 4=user causing corrupted DB; 8=count of users
          Options:=2;
          UserCount:=LDBUser_GetUsers(PSafeArray(TVarData(UserBuffer).VArray), PChar(DBName), Options);
          if UserCount < 0 then
            List.Add(LDBUser_GetError(UserCount))
          else
          begin
            for i:=1 to UserCount do
              List.Add(PChar(TVarData(UserBuffer[i]).VPointer));
            Result:=true;
          end;
        end;
      finally
        FreeLibrary(DLLHandle);
      end;
  finally
    NCursor;
  end;
end;

// tables
procedure GetTableList(DBName, Pwd: string; List: TStrings);
var
  db: Database;
begin
  db:=OpenMDB(DBName, Pwd);
  GetTableList(db, List);
end;

procedure GetTableList(db: Database; List: TStrings);
var
  i: integer;
  s: string;
begin
  List.Clear;
  for i:=0 to db.TableDefs.Count - 1 do
  begin
    s:=db.TableDefs[i].Name;
    if AnsiPosEx('MSYS', s, 1, false) <> 1 then
      List.Add(s);
  end;
end;

function GetTableFields(DBName, Pwd, TableName: string; List: TStrings): TFieldRecArray;
var
  db: Database;
begin
  db:=OpenMDB(DBName, Pwd);
  Result:=GetTableFields(db, TableName, List);
end;

function GetTableFields(db: Database; TableName: string; List: TStrings): TFieldRecArray;
var
  td: TableDef;
  i: integer;
begin
  Result:=nil;
  if Assigned(List) then
    List.Clear;
  td:=db.TableDefs[TableName];
  if not Assigned(td) then
    Exit;
  SetLength(Result, td.Fields.Count);
  for i:=0 to td.Fields.Count - 1 do
  begin
    Result[i].FieldName:=td.Fields[i].Name;
    Result[i].FieldType:=td.Fields[i].Type_;
    Result[i].FieldSize:=td.Fields[i].Size;
    Result[i].Attributes:=td.Fields[i].Attributes;
    Result[i].Required:=td.Fields[i].Required;
    Result[i].AllowZeroLength:=td.Fields[i].AllowZeroLength;
    Result[i].DefaultValue:=td.Fields[i].DefaultValue;
    Result[i].ValidationText:=td.Fields[i].ValidationText;
    Result[i].ValidationRule:=td.Fields[i].ValidationRule;
    if Assigned(List) then
      List.Add(td.Fields[i].Name);
  end;
end;

function GetTableIndexes(DBName, Pwd, TableName: string; NotRelations: boolean; List: TStrings): TIndexRecArray;
var
  db: Database;
begin
  db:=OpenMDB(DBName, Pwd);
  Result:=GetTableIndexes(db, TableName, NotRelations, List);
end;

function GetTableIndexes(db: Database; TableName: string; NotRelations: boolean; List: TStrings): TIndexRecArray;
var
  td: TableDef;
  i, j: integer;
  fn: string;
  Rel: TStrings;
begin
  Result:=nil;
  if Assigned(List) then
    List.Clear;
  td:=db.TableDefs[TableName];
  if not Assigned(td) then
    Exit;
  SetLength(Result, td.Indexes.Count);
  j:=0;
  Rel:=TStringList.Create;
  try
    GetRelations(db, Rel);
    for i:=0 to td.Indexes.Count - 1 do
    begin
      // ??? field is in format +FieldName
      fn:=copy(td.Indexes[i].Fields, 2, 255);
      if (not NotRelations) or (Rel.IndexOf(td.Indexes[i].Name) < 0) then
      begin
        Result[j].IndexName:=td.Indexes[i].Name;
        Result[j].Primary:=td.Indexes[i].Primary;
        Result[j].Unique:=td.Indexes[i].Unique;
        Result[j].Required:=td.Indexes[i].Required;
        Result[j].FieldName:=fn;
        if Assigned(List) then
          List.Add(td.Indexes[i].Name);
        Inc(j);
      end;
    end;
    SetLength(Result, j);
  finally
    Rel.Clear;
  end;
end;

procedure DeleteIndexes(db: Database; TableName: string; LeftPrimaryKey: boolean);
var
  tb: TableDef;
  i: integer;
begin
  tb:=db.TableDefs[TableName];
  for i:=tb.Indexes.Count - 1 downto 0 do
  begin
    if not(LeftPrimaryKey and tb.Indexes[i].Primary) then
      try
        tb.Indexes.Delete(tb.Indexes[i].Name);
      except
      end;
  end;
  db.TableDefs.Refresh;
end;

function CreateTable(db: Database; TableName: string; FieldRecArray: TFieldRecArray; IndexRecArray: TIndexRecArray): TableDef;
var
  tb: TableDef;
  fld: Field;
  idx: Index;
  i: integer;
begin
  tb:=db.CreateTableDef(TableName, 0, '', '');
  for i:=0 to High(FieldRecArray) do
  begin
    fld:=tb.CreateField(FieldRecArray[i].FieldName, FieldRecArray[i].FieldType, FieldRecArray[i].FieldSize);
    fld.Set_Attributes(FieldRecArray[i].Attributes);
    fld.Set_Required(FieldRecArray[i].Required);
    fld.Set_DefaultValue(FieldRecArray[i].DefaultValue);
    // text file?
    if FieldRecArray[i].FieldType = 10 then
      fld.Set_AllowZeroLength(FieldRecArray[i].AllowZeroLength);
    fld.Set_ValidationText(FieldRecArray[i].ValidationText);
    fld.Set_ValidationRule(FieldRecArray[i].ValidationRule);
    tb.Fields.Append(fld);
  end;
  for i:=0 to High(IndexRecArray) do
  begin
    idx:=tb.CreateIndex(IndexRecArray[i].IndexName);
    idx.Set_Primary(IndexRecArray[i].Primary);
    idx.Set_Unique(IndexRecArray[i].Unique);
    idx.Set_Required(IndexRecArray[i].Required);
    fld:=idx.CreateField(IndexRecArray[i].FieldName, EmptyParam, EmptyParam);
    idx.Fields.Append(fld);
    tb.Indexes.Append(idx);
  end;
  db.TableDefs.Append(tb);
  Result:=tb;
end;

procedure DropTable(db: Database; TableName: string);
begin
  db.TableDefs.Delete(TableName);
  db.TableDefs.Refresh;
end;

procedure RenameTable(db: Database; SrcTableName, DstTableName: string);
var
  tb: TableDef;
begin
  tb:=db.TableDefs[SrcTableName];
  tb.Set_Name(DstTableName);
  db.TableDefs.Refresh;
end;

// queries
function GetQueryDefs(DBName, Pwd: string; List: TStrings): TQueryDefArray;
var
  db: Database;
begin
  db:=OpenMDB(DBName, Pwd);
  Result:=GetQueryDefs(db, List);
end;

function GetQueryDefs(db: Database; List: TStrings): TQueryDefArray;
var
  i: integer;
begin
  if Assigned(List) then
    List.Clear;
  SetLength(Result, db.QueryDefs.Count);
  for i:=0 to db.QueryDefs.Count - 1 do
  begin
    Result[i].Name:=db.QueryDefs[i].Name;
    Result[i].SQL:=db.QueryDefs[i].SQL;
    if Assigned(List) then
      List.Add(db.QueryDefs[i].Name);
  end;
end;

procedure ClearQueryDefs(db: Database);
var
  i: integer;
begin
  for i:=db.QueryDefs.Count - 1 downto 0 do
    db.QueryDefs.Delete(db.QueryDefs[i].Name);
  db.QueryDefs.Refresh;
end;

function CreateQueryDef(db: Database; QDef: TQueryDef): QueryDef;
begin
  Result:=db.CreateQueryDef(QDef.Name, QDef.SQL);
end;

procedure CreateQueryDefs(db: Database; QDefArr: TQueryDefArray);
var
  i: integer;
begin
  for i:=0 to High(QDefArr) do
    db.CreateQueryDef(QDefArr[i].Name, QDefArr[i].SQL);
  db.QueryDefs.Refresh;
end;

// relations
function GetRelations(DBName, Pwd: string; List: TStrings): TRelationRecArray;
var
  db: Database;
begin
  db:=OpenMDB(DBName, Pwd);
  Result:=GetRelations(db, List);
end;

function GetRelations(db: Database; List: TStrings): TRelationRecArray;
var
  Rel: Relation;
  i, j: integer;
begin
  if Assigned(List) then
    List.Clear;
  SetLength(Result, db.Relations.Count);
  for i:=0 to db.Relations.Count - 1 do
  begin
    Rel:=db.Relations[i];
    Result[i].Name:=Rel.Name;
    Result[i].Table:=Rel.Table;
    Result[i].ForeignTable:=Rel.ForeignTable;
    Result[i].Attributes:=Rel.Attributes;
    SetLength(Result[i].Fields, Rel.Fields.Count);
    for j:=0 to Rel.Fields.Count - 1 do
    begin
      Result[i].Fields[j].FieldName:=Rel.Fields[j].Name;
      Result[i].Fields[j].ForeignName:=Rel.Fields[j].ForeignName;
    end;
    if Assigned(List) then
      List.Add(Rel.Name);
  end;
end;

procedure CreateRelations(db: Database; RelationRecArray: TRelationRecArray);
var
  Rel: Relation;
  fld: Field;
  i, j: integer;
begin
  for i:=0 to High(RelationRecArray) do
  begin
    Rel:=db.CreateRelation(RelationRecArray[i].Name, RelationRecArray[i].Table, RelationRecArray[i].ForeignTable, RelationRecArray[i].Attributes);
    for j:=0 to High(RelationRecArray[i].Fields) do
    begin
      fld:=Rel.CreateField(RelationRecArray[i].Fields[j].FieldName, EmptyParam, EmptyParam);
      fld.Set_ForeignName(RelationRecArray[i].Fields[j].ForeignName);
      Rel.Fields.Append(fld);
    end;
    db.Relations.Append(Rel);
  end;
end;

procedure ClearRelations(db: Database);
var
  i: integer;
begin
  for i:=db.Relations.Count - 1 downto 0 do
    db.Relations.Delete(db.Relations.Item[i].Name);
end;

end.
